#!/usr/bin/python3
"""Check for available updates using the client dbus interface.

More information at:
https://wiki.ubuntu.com/ImageBasedUpgrades/Client#DBus_API

"""
import dbus
import logging

from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib


def main():
    bus = dbus.SystemBus()
    service = bus.get_object('com.canonical.SystemImage', '/Service')
    iface = dbus.Interface(service, 'com.canonical.SystemImage')

    auto_download = int(iface.GetSetting('auto_download'))
    logging.info('GetSetting(auto_download) = {!r}'.format(auto_download))
    assert auto_download == 1

    def check_for_update():
        logging.info('Calling CheckForUpdate...')
        iface.CheckForUpdate()

    def download_update():
        logging.info('Calling DownloadUpdate...')
        iface.DownloadUpdate()

    def apply_update():
        logging.info('Calling ApplyUpdate...')
        iface.ApplyUpdate()

    def update_available_status_cb(*args, **kwargs):
        logging.debug(
            'UpdateAvailableStatus signal received:\nargs: {}\nkwargs:{}'
            .format(args, kwargs))

        is_available = bool(args[0])
        assert is_available

        downloading = bool(args[1])
        assert downloading

        available_version = args[2]
        update_size = args[3]
        last_update_date = args[4]

        # TBD: Update when bug#1230263 is fixed
        # descriptions = args[5]
        # error_reason = args[6]
        error_reason = args[5]
        assert not error_reason

        logging.info(
            'Update available: (version: {}, size: {}, date: {}'
            .format(available_version, update_size, last_update_date))
        download_update()

    def update_progress_cb(percentage, eta):
        logging.debug(
            'UpdateProgress signal received: (percentage: {}, eta: {}'
            .format(percentage, eta))

    def update_downloaded_cb():
        logging.debug('Update downloaded')
        apply_update()

    iface.connect_to_signal(
        'UpdateAvailableStatus',
        update_available_status_cb)
    iface.connect_to_signal(
        'UpdateProgress',
        update_progress_cb)
    iface.connect_to_signal(
        'UpdateDownloaded',
        update_downloaded_cb)

    GLib.timeout_add(2000, check_for_update)
    loop = GLib.MainLoop()
    loop.run()


if __name__ == '__main__':
    DBusGMainLoop(set_as_default=True)
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s %(levelname)s: %(message)s')
    main()
