#!/usr/bin/python
"""Get revision number needed to force a full upgrade."""

import argparse
import json
import logging
import os
import sys
import urllib2


def main(args):
    """Find version for which delta is more expensive than full upgrade.

    :param args: Arguments passed from the command line already parsed
    :type args: argparse.Namespace
    :returns: None

    """

    # Get index.json file for the channel and device requested
    base_url = 'http://system-image.ubuntu.com'
    url = os.path.join(base_url, args.channel, args.device, 'index.json')
    logging.info('Downloading {}...'.format(url))
    response = urllib2.urlopen(url)
    assert response.code == 200
    index_data = json.load(response)
    images = index_data['images']

    # Get version span
    full_images = [image for image in images if image['type'] == 'full']
    delta_images = [image for image in images if image['type'] == 'delta']

    oldest_image = full_images[0]
    oldest_version = oldest_image['version']
    logging.info('Oldest version: {}'.format(oldest_version))
    latest_image = full_images[-1]
    latest_version = latest_image['version']
    logging.info('Latest version: {}'.format(latest_version))
    full_download_size = sum(f['size'] for f in latest_image['files'])

    for version in range(latest_version - 1, oldest_version - 1, -1):
        # Find delta download size from an older version to the latest one
        delta_download_size = get_delta_download_size(
            delta_images, version, latest_version, [])

        # Stop if delta download is more expensive than full download
        if delta_download_size and delta_download_size > full_download_size:
            logging.info(
                'Delta download size ({}) bigger than full download size ({})'
                .format(delta_download_size, full_download_size))
            print version
            return 0
    else:
        logging.error(
            'No delta download size found bigger than full download size ({})'
            .format(full_download_size))
        return 1


def get_delta_download_size(delta_images, from_version, to_version, path):
    """Get delta download size from a version to another one.

    :param delta_images: All delta images data
    :type delta_images: list(dict)
    :param from_version: Initial version in the upgrade path
    :type from_version: int
    :param to_version: Final version in the upgrade path
    :type to_version: int
    :param path: Path for the ongoing upgrade calculation
    :type path: list(dict)
    :returns: Download size
    :rtype: int

    """
    logging.debug('Getting delta download size {} -> {}'
                  .format(from_version, to_version))

    if from_version == to_version:
        versions = [image['base'] for image in path]
        versions.append(path[-1]['version'])
        assert path[-1]['version'] == to_version
        logging.info(
            'Upgrade path: {}'
            .format(' -> '.join(str(version) for version in versions)))

        delta_download_size = sum(sum(f['size'] for f in image['files'])
                                  for image in path)
        logging.info('Delta download size ({} -> {}): {}'
                     .format(versions[0], versions[-1], delta_download_size))
        return delta_download_size
    else:
        candidate_images = [image for image in delta_images
                            if (image['base'] == from_version
                                and image['version'] <= to_version)]
        if not candidate_images:
            logging.warning('No candidate images found ({} -> {})'
                            .format(from_version, to_version))
            return None

        # It's assumed that the delta image chosen is the one that updates to
        # the newer version
        candidate_image = sorted(
            candidate_images, key=lambda img: img['version'], reverse=True)[0]
        logging.debug('Candidate image: {} -> {}'
                      .format(candidate_image['base'],
                              candidate_image['version']))
        path.append(candidate_image)
        return get_delta_download_size(
            delta_images, candidate_image['version'], to_version, path)


def parse_cmdline():
    """Parse command line arguments.

    :returns: Parsed arguments
    :rtype: argparse.Namespace

    """
    parser = argparse.ArgumentParser(
        description=__doc__,
        epilog='Example: %(prog)s devel-proposed mako',
    )
    parser.add_argument('channel')
    parser.add_argument('device')
    return parser.parse_args()

if __name__ == '__main__':
    logging.basicConfig(
        format='%(levelname)s: %(message)s',
        level=logging.DEBUG,
    )
    args = parse_cmdline()
    sys.exit(main(args))
