#!/usr/bin/python3
"""Quick & dirty delta upgrade test setup."""

import fnmatch
import hashlib
import json
import logging
import os
import shutil
import subprocess
import tarfile
import tempfile
import time
import urllib.request

from os.path import join, basename, dirname, relpath

import gnupg


class UpgradeFiles(object):
    CHANNELS_FNAME = 'channels.json'
    INDEX_FNAME = 'index.json'

    CHANNELS = ['devel-proposed']
    DEVICES = ['mako']

    OFFICIAL_SERVER = 'http://system-image.ubuntu.com'

    def __init__(self, channel, device):
        self.channel = channel
        self.device = device
        assert channel in self.CHANNELS
        assert device in self.DEVICES

        logging.info('Creating server files directory...')
        self.directory = tempfile.mkdtemp(prefix='upgrade-test-')
        CLEANUP.add_directory(self.directory)

        self.keyring = {}
        self._configure_keyrings()
        self._configure_files()

    def _configure_keyrings(self):
        logging.info('Creating gpg directory...')
        gpg_dir = join(self.directory, 'gpg')
        os.mkdir(gpg_dir)

        # Archive master keyring
        amk = Keyring('archive-master')
        amk.sign_keyring(amk)
        amk.export(gpg_dir)
        assert amk.verify(
            amk.get_abspath(amk.asc_path),
            amk.get_abspath(amk.KEYRING_FNAME))

        # Image master keyring
        imk = Keyring('image-master')
        amk.sign_keyring(imk)
        imk.export(gpg_dir)
        assert amk.verify(
            imk.get_abspath(imk.asc_path),
            imk.get_abspath(imk.KEYRING_FNAME))

        # Image signing keyring
        isk = Keyring('image-signing')
        imk.sign_keyring(isk)
        isk.export(gpg_dir)
        assert imk.verify(
            isk.get_abspath(isk.ask_path),
            isk.get_abspath(isk.KEYRING_FNAME))

        # Keep track of keyring objects to sign other files
        self.keyring['archive-master'] = amk
        self.keyring['image-master'] = imk
        self.keyring['image-signing'] = isk

    def _configure_files(self):
        self._configure_channels_file()

        for channel in self.CHANNELS:
            for device in self.DEVICES:
                self._configure_index_file(
                    channel,
                    device,
                )

    def _configure_channels_file(self):
        logging.info('Creating {}...'.format(self.CHANNELS_FNAME))
        channels_data = {channel: {
            device: {'index': ('/{}/{}/{}'
                               .format(channel, device, self.INDEX_FNAME))}
            for device in self.DEVICES}
            for channel in self.CHANNELS
        }
        channels_path = join(self.directory, self.CHANNELS_FNAME)
        with open(channels_path, 'w') as fp:
            json.dump(channels_data, fp, sort_keys=True, indent=4)

        self.keyring['image-signing'].sign(channels_path)

    def _configure_index_file(self, channel, device):
        device_dir = join(self.directory, channel, device)
        os.makedirs(device_dir)

        # Get file from original server
        index_url = join(
            self.OFFICIAL_SERVER,
            channel,
            device,
            self.INDEX_FNAME)

        logging.info('Opening {}...'.format(index_url))
        response = urllib.request.urlopen(index_url)
        assert response.code == 200
        response_data = response.read().decode('utf-8')

        image_signing_keyring = self.keyring['image-signing']
        index_data = json.loads(response_data)
        if channel == self.channel and device in self.device:
            self._configure_delta_file(
                index_data,
                device_dir,
                image_signing_keyring,
            )

        index_path = join(device_dir, self.INDEX_FNAME)
        with open(index_path, 'w') as fp:
            json.dump(index_data, fp, sort_keys=True, indent=4)
        image_signing_keyring.sign(index_path)

    def _configure_delta_file(self, index_data, device_dir,
                              image_signing_keyring):
        delta_dir = tempfile.mkdtemp(prefix='upgrade-test-delta-')
        CLEANUP.add_directory(delta_dir)
        etc_dir = join(delta_dir, 'system', 'etc')
        os.makedirs(etc_dir)
        test_path = join(etc_dir, 'upgrade-test')
        with open(test_path, 'w') as fp:
            fp.write('Upgrade test')

        delta_fname = 'upgrade.tar.xz'
        delta_path = join(device_dir, delta_fname)
        with tarfile.open(delta_path, 'w:xz') as fp:
            fp.add(join(delta_dir, 'system'), 'system')

        with open(delta_path, 'rb') as fp:
            sha256 = hashlib.sha256()
            sha256.update(fp.read())
            checksum = sha256.hexdigest()

        delta_signature_path = \
            image_signing_keyring.sign(delta_path)

        images = index_data['images']
        last_full_image = next(
            image
            for image in reversed(images)
            if image['type'] == 'full'
        )
        last_version = last_full_image['version']

        index_data['images'].append({
            'description': 'UpgradeTest',
            'files': [{
                'checksum': checksum,
                'order': 0,
                'path': '/' + relpath(delta_path, self.directory),
                'signature': (
                    '/' + relpath(delta_signature_path, self.directory)),
                'size': os.path.getsize(delta_path),
            },
            ],
            'type': 'delta',
            'base': last_version,
            'version': last_version + 1,
        })


class Path(object):
    def __init__(self, path):
        self.path = path

    def push(self, directory):
        if os.path.isdir(self.path):
            destination = directory
        else:
            destination = join(directory,
                               basename(self.path))
        logging.info('Pushing {} to {}...'.format(self.path, destination))
        Command('adb push {} {}'.format(self.path, destination)).run()


class Keyring(object):

    KEYRING_FNAME = 'keyring.gpg'
    KEYRING_JSON_FNAME = 'keyring.json'

    def __init__(self, keyring_type):
        self.keyring_type = keyring_type
        self.tar_path = None
        self.asc_path = None

        logging.info('Creating {} keyring directory...'.format(keyring_type))
        self.keyring_dir = tempfile.mkdtemp(prefix='upgrade-test-keyring-')
        CLEANUP.add_directory(self.keyring_dir)

        logging.info('Creating {} keyring file...'.format(keyring_type))
        self.gpg = gnupg.GPG(
            gnupghome=self.keyring_dir,
            keyring=self.KEYRING_FNAME)

        logging.info('Creating {} keyring json file...'.format(keyring_type))
        keyring_json_path = self.get_abspath(self.KEYRING_JSON_FNAME)
        with open(keyring_json_path, 'w') as fp:
            json.dump({'type': keyring_type}, fp)

        logging.info('Generating {} key (this might take a while)...'
                     .format(keyring_type))
        self.key = self.gpg.gen_key(self.gpg.gen_key_input())

        self._create_tar_file()

    def _create_tar_file(self):
        logging.info('Creating {} tar file..'.format(self.keyring_type))
        keyring_path = self.get_abspath(self.KEYRING_FNAME)
        keyring_json_path = self.get_abspath(self.KEYRING_JSON_FNAME)

        keyring_tar_fname = '{}.tar.xz'.format(self.keyring_type)
        self.tar_path = self.get_abspath(keyring_tar_fname)

        logging.info('Creating {}...'.format(self.tar_path))
        with tarfile.open(self.tar_path, 'w:xz') as fp:
            for abspath in [keyring_path, keyring_json_path]:
                logging.info('Adding {} to the tar file...'.format(abspath))
                fp.add(abspath, basename(abspath))

    def get_abspath(self, rel_path):
        return join(self.keyring_dir, rel_path)

    def sign(self, abspath):
        logging.info('Signing {} with {} key...'
                     .format(abspath, self.keyring_type))
        with open(abspath, 'rb') as fp:
            signature = str(self.gpg.sign_file(fp, detach=True))

        signed_abspath = '{}.asc'.format(abspath)
        logging.info('Writing signature to {}...'.format(signed_abspath))
        with open(signed_abspath, 'w') as fp:
            fp.write(signature)

        return signed_abspath

    def verify(self, signed_abspath, abspath):
        logging.info('Verifying signature {} for {} with {} key...'
                     .format(signed_abspath, abspath, self.keyring_type))
        with open(signed_abspath, 'rb') as fp:
            verified = self.gpg.verify_file(fp, abspath)

        return verified

    def sign_keyring(self, other_keyring):
        logging.info('Signing {} keyring with {} key'
                     .format(other_keyring.keyring_type, self.keyring_type))
        other_keyring.asc_path = self.sign(other_keyring.tar_path)

    def export(self, target_dir):
        logging.info('Exporting {} keyring...'.format(self.keyring_type))
        shutil.copy(self.tar_path, target_dir)
        shutil.copy(self.asc_path, target_dir)


class Command(object):
    """A wrapper around subprocess.Popen with other defaults and logging."""
    def __init__(self, cmd, **kwargs):
        self.cmd = cmd
        if isinstance(self.cmd, str):
            self.cmd_str = self.cmd
        elif isinstance(self.cmd, list):
            self.cmd_str = ' '.join(self.cmd)
        else:
            raise TypeError('cmd should be either <str> or <list>')

        kwargs.setdefault('stdout', subprocess.PIPE)
        kwargs.setdefault('stderr', subprocess.PIPE)
        kwargs.setdefault('shell', True)
        self.kwargs = kwargs

    def run(self, background=False):
        logging.info('Running {!r}...'.format(self.cmd_str))
        process = subprocess.Popen(self.cmd, **self.kwargs)

        if not background:
            stdout, stderr = process.communicate()
            self.stdout = stdout.decode('utf-8') if stdout else None
            self.stderr = stderr.decode('utf-8') if stderr else None
            if process.returncode == 0:
                log_level = logging.DEBUG
            else:
                log_level = logging.ERROR

            message = ['Return code: {}'.format(process.returncode)]
            if self.stdout:
                message.append('stdout:\n{}'.format(self.stdout))
            if self.stderr:
                message.append('stderr:\n{}'.format(self.stderr))
            message = '\n'.join(message)

            logging.log(log_level, message)

        return self


class Cleanup(object):
    def __init__(self):
        self.directories = []

    def add_directory(self, directory):
        self.directories.append(directory)

    def run(self):
        pass
        #for directory in self.directories:
            #logging.info('Removing directory {}...'.format(directory))
            #shutil.rmtree(directory)

CLEANUP = Cleanup()


def flash_image():
    """Flash device with latest image."""
    Command('phablet-flash ubuntu-system '
            '--channel devel-proposed --no-backup').run()
    logging.debug('Waiting for device to boot up...')
    Command('adb wait-for-device').run()
    time.sleep(60)
    logging.debug('Configuring network interface...')
    Command('phablet-network').run()
    time.sleep(60)

    logging.debug('Making sure image is writable...')
    Command('adb shell touch /userdata/.writable_image').run()
    Command('adb shell reboot').run()
    logging.debug('Waiting for device to boot up...')
    Command('adb wait-for-device').run()
    time.sleep(60)
    logging.debug('Installing system-image-dev package...')
    Command('adb shell apt-get install system-image-dev').run()


def get_image_info():
    """Get image info"""
    logging.debug('Getting image info...')
    cmd = Command('adb shell system-image-cli -i').run()
    lines = cmd.stdout.splitlines()
    info = dict([line.split(': ', 1) for line in lines])
    channel = info['channel']
    device = info['device name']
    logging.debug('channel: {}, device: {}'.format(channel, device))
    return channel, device


def flash_recovery_partition(upgrade_files):
    """Flash recovery partition with test keys."""

    # TBD: This directory should be configurable and/or depend on current user
    # name
    download_dir = '/var/lib/jenkins/phablet-flash/imageupdates/pool'
    logging.debug('Looking for device files in {}...'.format(download_dir))
    files = os.listdir(download_dir)
    device_files = sorted(
        fnmatch.filter(files, 'device-*.tar.xz'),
        key=lambda f: os.path.getmtime(join(download_dir, f)),
        reverse=True,
    )
    device_file = join(download_dir, device_files[0])
    logging.debug('Device file: {}'.format(device_file))

    recovery_dir = tempfile.mkdtemp(prefix='upgrade-test-recovery-')
    CLEANUP.add_directory(recovery_dir)

    # Locate latest image file
    recovery_fname = 'recovery.img'
    with tarfile.open(device_file) as tf:
        logging.debug('Extracting recovery.img...')
        tf.extract(join('partitions', recovery_fname), recovery_dir)

    recovery_path = join(recovery_dir, 'partitions', recovery_fname)

    logging.debug('Extracting initrd...')
    Command(
        'abootimg -x {} cmdline kernel initrd'.format(recovery_path),
        cwd=recovery_dir,
    ).run()

    logging.debug('Updating image size...')
    Command(
        "sed -ri 's/bootsize = 0x[0-9]+/bootsize = 0xa00000/' cmdline",
        cwd=recovery_dir,
    ).run()

    logging.debug('Extracting archive master keyring...')
    cpio_dir = join(recovery_dir, 'cpio')
    os.mkdir(cpio_dir)
    Command(
        'cat ../initrd | gzip -d | cpio -i',
        cwd=cpio_dir,
    ).run()

    logging.debug('Copying *testing* archive master keyring...')
    archive_master_keyring = upgrade_files.keyring['archive-master']
    target_dir = join(cpio_dir, 'etc', 'system-image')
    shutil.copy(archive_master_keyring.tar_path, target_dir)
    shutil.copy(archive_master_keyring.asc_path, target_dir)

    logging.debug('Repacking initrd...')
    with open(join(recovery_dir, 'initrd'), 'wb') as fd:
        Command(
            'find . | cpio -o --format=newc | gzip -9',
            stdout=fd,
            cwd=cpio_dir,
        ).run()

    logging.debug('Repacking recovery.img...')
    Command(
        'abootimg -u {} -f cmdline'.format(recovery_path),
        cwd=recovery_dir,
    ).run()
    Command(
        'abootimg -u {} -r initrd'.format(recovery_path),
        cwd=recovery_dir,
    ).run()

    logging.debug('Flashing test recovery image...')
    Command('adb shell reboot -f bootloader').run()
    Command('fastboot flash recovery {}'.format(recovery_path)).run()
    Command('fastboot reboot').run()

    logging.debug('Waiting for device to boot up...')
    Command('adb wait-for-device').run()
    time.sleep(60)


def push_files(upgrade_files):
    """Push upgrade test files to device."""
    base = dirname(__file__)
    target_base = '/tmp/upgrade-test'
    Path(join(base, 'scripts')).push(join(target_base, 'scripts'))
    Path(join(base, 'data')).push(join(target_base, 'data'))
    Path(upgrade_files.directory).push(join(target_base, 'server-files'))

    logging.debug(
        'Copying generated archive master key to default location...')
    Command('adb shell cp '
            '/tmp/upgrade-test/server-files/gpg/archive-master.tar.* '
            '/etc/system-image').run()

    logging.debug('Patching ubuntu-download DBus service file')
    service_file = (
        '/usr/share/dbus-1/system-services/'
        'com.canonical.applications.Downloader.service')
    Command(
        r"adb shell sed -ri 's|^Exec=.*$|\0 -self-signed-certs "
        r"/tmp/upgrade-test/data/*.pem|' {}"
        .format(service_file),
    ).run()


def main():
    #flash_image()
    channel, device = get_image_info()

    upgrade_files = UpgradeFiles(channel, device)
    flash_recovery_partition(upgrade_files)

    # Push upgrade files after recovery partition is flashed
    # (otherwise files are removed after reboot)
    push_files(upgrade_files)

    logging.debug('Launching HTTP/HTTPS server...')
    (Command('adb shell /tmp/upgrade-test/scripts/server.py')
     .run(background=True))

    logging.debug('Launching client DBus service...')
    (Command('adb shell system-image-dbus '
             '-C /tmp/upgrade-test/data/client.ini '
             '--testing=live -v').run(background=True))

    logging.debug('Using DBus API to get upgrade from server...')
    (Command('adb shell /tmp/upgrade-test/scripts/check-updates.py')
     .run(background=True))

    logging.debug('Waiting for device to reboot...')
    time.sleep(60)
    Command('adb wait-for-device').run()
    time.sleep(60)

    logging.debug('Checking that upgrade was correctly applied')
    stdout = Command('adb shell cat /etc/upgrade-test').run().stdout
    assert stdout == 'Upgrade test'


if __name__ == '__main__':
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s %(levelname)s: %(message)s')
    try:
        main()
    finally:
        CLEANUP.run()
